@@ -1,32 +1,58 @@
 # -*- test-case-name: twisted.names.test.test_dns -*-
-# Copyright (c) 2001-2007 Twisted Matrix Laboratories.
+# Copyright (c) Twisted Matrix Laboratories.
 # See LICENSE for details.
 
-
 """
 DNS protocol implementation.
 
 Future Plans:
     - Get rid of some toplevels, maybe.
-    - Put in a better lookupRecordType implementation.
 
-@author: U{Moshe Zadka<mailto:moshez@twistedmatrix.com>},
-         U{Jp Calderone<mailto:exarkun@twistedmatrix.com>}
+@author: Moshe Zadka
+@author: Jean-Paul Calderone
 """
 
+__all__ = [
+    'IEncodable', 'IRecord',
+
+    'A', 'A6', 'AAAA', 'AFSDB', 'CNAME', 'DNAME', 'HINFO',
+    'MAILA', 'MAILB', 'MB', 'MD', 'MF', 'MG', 'MINFO', 'MR', 'MX',
+    'NAPTR', 'NS', 'NULL', 'PTR', 'RP', 'SOA', 'SPF', 'SRV', 'TXT', 'WKS',
+
+    'ANY', 'CH', 'CS', 'HS', 'IN',
+
+    'ALL_RECORDS', 'AXFR', 'IXFR',
+
+    'EFORMAT', 'ENAME', 'ENOTIMP', 'EREFUSED', 'ESERVER',
+
+    'Record_A', 'Record_A6', 'Record_AAAA', 'Record_AFSDB', 'Record_CNAME',
+    'Record_DNAME', 'Record_HINFO', 'Record_MB', 'Record_MD', 'Record_MF',
+    'Record_MG', 'Record_MINFO', 'Record_MR', 'Record_MX', 'Record_NAPTR',
+    'Record_NS', 'Record_NULL', 'Record_PTR', 'Record_RP', 'Record_SOA',
+    'Record_SPF', 'Record_SRV', 'Record_TXT', 'Record_WKS', 'UnknownRecord',
+
+    'QUERY_CLASSES', 'QUERY_TYPES', 'REV_CLASSES', 'REV_TYPES', 'EXT_QUERIES',
+
+    'Charstr', 'Message', 'Name', 'Query', 'RRHeader', 'SimpleRecord',
+    'DNSDatagramProtocol', 'DNSMixin', 'DNSProtocol',
+
+    'OK', 'OP_INVERSE', 'OP_NOTIFY', 'OP_QUERY', 'OP_STATUS', 'OP_UPDATE',
+    'PORT',
+
+    'AuthoritativeDomainError', 'DNSQueryTimeoutError', 'DomainError',
+    ]
+
+
 # System imports
 import warnings
 
 import struct, random, types, socket
 
-try:
-    import cStringIO as StringIO
-except ImportError:
-    import StringIO
+import cStringIO as StringIO
 
 AF_INET6 = socket.AF_INET6
 
-from zope.interface import implements, Interface
+from zope.interface import implements, Interface, Attribute
 
 
 # Twisted imports
@@ -50,8 +76,10 @@ PORT = 53
  RP, AFSDB) = range(1, 19)
 AAAA = 28
 SRV = 33
+NAPTR = 35
 A6 = 38
 DNAME = 39
+SPF = 99
 
 QUERY_TYPES = {
     A: 'A',
@@ -77,9 +105,10 @@ QUERY_TYPES = {
 
     AAAA: 'AAAA',
     SRV: 'SRV',
-
+    NAPTR: 'NAPTR',
     A6: 'A6',
-    DNAME: 'DNAME'
+    DNAME: 'DNAME',
+    SPF: 'SPF'
 }
 
 IXFR, AXFR, MAILB, MAILA, ALL_RECORDS = range(251, 256)
@@ -124,10 +153,10 @@ OK, EFORMAT, ESERVER, ENAME, ENOTIMP, EREFUSED = range(6)
 class IRecord(Interface):
     """
     An single entry in a zone of authority.
-
-    @cvar TYPE: An indicator of what kind of record this is.
     """
 
+    TYPE = Attribute("An indicator of what kind of record this is.")
+
 
 # Backwards compatibility aliases - these should be deprecated or something I
 # suppose. -exarkun
@@ -164,6 +193,7 @@ class IEncodable(Interface):
     Interface for something which can be encoded to and decoded
     from a file object.
     """
+
     def encode(strio, compDict = None):
         """
         Write a representation of this object to the given
@@ -194,6 +224,61 @@ class IEncodable(Interface):
         """
 
 
+
+class Charstr(object):
+    implements(IEncodable)
+
+    def __init__(self, string=''):
+        if not isinstance(string, str):
+            raise ValueError("%r is not a string" % (string,))
+        self.string = string
+
+
+    def encode(self, strio, compDict=None):
+        """
+        Encode this Character string into the appropriate byte format.
+
+        @type strio: file
+        @param strio: The byte representation of this Charstr will be written
+            to this file.
+        """
+        string = self.string
+        ind = len(string)
+        strio.write(chr(ind))
+        strio.write(string)
+
+
+    def decode(self, strio, length=None):
+        """
+        Decode a byte string into this Name.
+
+        @type strio: file
+        @param strio: Bytes will be read from this file until the full string
+            is decoded.
+
+        @raise EOFError: Raised when there are not enough bytes available from
+            C{strio}.
+        """
+        self.string = ''
+        l = ord(readPrecisely(strio, 1))
+        self.string = readPrecisely(strio, l)
+
+
+    def __eq__(self, other):
+        if isinstance(other, Charstr):
+            return self.string == other.string
+        return False
+
+
+    def __hash__(self):
+        return hash(self.string)
+
+
+    def __str__(self):
+        return self.string
+
+
+
 class Name:
     implements(IEncodable)
 
@@ -244,7 +329,11 @@ class Name:
 
         @raise EOFError: Raised when there are not enough bytes available
         from C{strio}.
+
+        @raise ValueError: Raised when the name cannot be decoded (for example,
+            because it contains a loop).
         """
+        visited = set()
         self.name = ''
         off = 0
         while 1:
@@ -256,6 +345,9 @@ class Name:
             if (l >> 6) == 3:
                 new_off = ((l&63) << 8
                             | ord(readPrecisely(strio, 1)))
+                if new_off in visited:
+                    raise ValueError("Compression loop in encoded name")
+                visited.add(new_off)
                 if off == 0:
                     off = strio.tell()
                 strio.seek(new_off)
@@ -342,7 +434,7 @@ class Query:
         return 'Query(%r, %r, %r)' % (str(self.name), self.type, self.cls)
 
 
-class RRHeader:
+class RRHeader(tputil.FancyEqMixin):
     """
     A resource record header.
 
@@ -353,11 +445,15 @@ class RRHeader:
     @ivar cls: The query class of the original request.
     @ivar ttl: The time-to-live for this record.
     @ivar payload: An object that implements the IEncodable interface
-    @ivar auth: Whether this header is authoritative or not.
+
+    @ivar auth: A C{bool} indicating whether this C{RRHeader} was parsed from an
+        authoritative message.
     """
 
     implements(IEncodable)
 
+    compareAttributes = ('name', 'type', 'cls', 'ttl', 'payload', 'auth')
+
     fmt = "!HHIH"
 
     name = None
@@ -386,7 +482,7 @@ class RRHeader:
         @type payload: An object implementing C{IEncodable}
         @param payload: A Query Type specific data object.
         """
-        assert (payload is None) or (payload.TYPE == type)
+        assert (payload is None) or isinstance(payload, UnknownRecord) or (payload.TYPE == type)
 
         self.name = Name(name)
         self.type = type
@@ -473,6 +569,7 @@ class Record_NS(SimpleRecord):
     An authoritative nameserver.
     """
     TYPE = NS
+    fancybasename = 'NS'
 
 
 
@@ -485,6 +582,7 @@ class Record_MD(SimpleRecord):
     @see: L{Record_MX}
     """
     TYPE = MD
+    fancybasename = 'MD'
 
 
 
@@ -497,6 +595,7 @@ class Record_MF(SimpleRecord):
     @see: L{Record_MX}
     """
     TYPE = MF
+    fancybasename = 'MF'
 
 
 
@@ -505,6 +604,7 @@ class Record_CNAME(SimpleRecord):
     The canonical name for an alias.
     """
     TYPE = CNAME
+    fancybasename = 'CNAME'
 
 
 
@@ -515,6 +615,7 @@ class Record_MB(SimpleRecord):
     This is an experimental record type.
     """
     TYPE = MB
+    fancybasename = 'MB'
 
 
 
@@ -525,6 +626,7 @@ class Record_MG(SimpleRecord):
     This is an experimental record type.
     """
     TYPE = MG
+    fancybasename = 'MG'
 
 
 
@@ -535,6 +637,7 @@ class Record_MR(SimpleRecord):
     This is an experimental record type.
     """
     TYPE = MR
+    fancybasename = 'MR'
 
 
 
@@ -543,6 +646,7 @@ class Record_PTR(SimpleRecord):
     A domain name pointer.
     """
     TYPE = PTR
+    fancybasename = 'PTR'
 
 
 
@@ -555,8 +659,10 @@ class Record_DNAME(SimpleRecord):
     maps a single node of the name space.
 
     @see: U{http://www.faqs.org/rfcs/rfc2672.html}
+    @see: U{http://www.faqs.org/rfcs/rfc3363.html}
     """
     TYPE = DNAME
+    fancybasename = 'DNAME'
 
 
 
@@ -598,7 +704,8 @@ class Record_A(tputil.FancyEqMixin):
 
 
     def __str__(self):
-        return '<A %s ttl=%s>' % (self.dottedQuad(), self.ttl)
+        return '<A address=%s ttl=%s>' % (self.dottedQuad(), self.ttl)
+    __repr__ = __str__
 
 
     def dottedQuad(self):
@@ -647,7 +754,8 @@ class Record_SOA(tputil.FancyEqMixin, tputil.FancyStrMixin):
     """
     implements(IEncodable, IRecord)
 
-    compareAttributes = ('serial', 'mname', 'rname', 'refresh', 'expire', 'retry', 'ttl')
+    fancybasename = 'SOA'
+    compareAttributes = ('serial', 'mname', 'rname', 'refresh', 'expire', 'retry', 'minimum', 'ttl')
     showAttributes = (('mname', 'mname', '%s'), ('rname', 'rname', '%s'), 'serial', 'refresh', 'retry', 'expire', 'minimum', 'ttl')
 
     TYPE = SOA
@@ -688,7 +796,7 @@ class Record_SOA(tputil.FancyEqMixin, tputil.FancyStrMixin):
 
 
 
-class Record_NULL:                   # EXPERIMENTAL
+class Record_NULL(tputil.FancyStrMixin, tputil.FancyEqMixin):
     """
     A null record.
 
@@ -700,6 +808,9 @@ class Record_NULL:                   # EXPERIMENTAL
     """
     implements(IEncodable, IRecord)
 
+    fancybasename = 'NULL'
+    showAttributes = compareAttributes = ('payload', 'ttl')
+
     TYPE = NULL
 
     def __init__(self, payload=None, ttl=None):
@@ -720,7 +831,7 @@ class Record_NULL:                   # EXPERIMENTAL
 
 
 
-class Record_WKS(tputil.FancyEqMixin, tputil.FancyStrMixin):                    # OBSOLETE
+class Record_WKS(tputil.FancyEqMixin, tputil.FancyStrMixin):
     """
     A well known service description.
 
@@ -744,11 +855,14 @@ class Record_WKS(tputil.FancyEqMixin, tputil.FancyStrMixin):
     """
     implements(IEncodable, IRecord)
 
+    fancybasename = "WKS"
     compareAttributes = ('address', 'protocol', 'map', 'ttl')
-    showAttributes = ('address', 'protocol', 'ttl')
+    showAttributes = [('_address', 'address', '%s'), 'protocol', 'ttl']
 
     TYPE = WKS
 
+    _address = property(lambda self: socket.inet_ntoa(self.address))
+
     def __init__(self, address='0.0.0.0', protocol=0, map='', ttl=None):
         self.address = socket.inet_aton(address)
         self.protocol, self.map = protocol, map
@@ -772,12 +886,10 @@ class Record_WKS(tputil.FancyEqMixin, tputil.FancyStrMixin):
 
 
 
-class Record_AAAA(tputil.FancyEqMixin):               # OBSOLETE (or headed there)
+class Record_AAAA(tputil.FancyEqMixin, tputil.FancyStrMixin):
     """
     An IPv6 host address.
 
-    This record type is obsolete.
-
     @type address: C{str}
     @ivar address: The packed network-order representation of the IPv6 address
         associated with this record.
@@ -786,13 +898,17 @@ class Record_AAAA(tputil.FancyEqMixin):               # OBSOLETE (or headed ther
     @ivar ttl: The maximum number of seconds which this record should be
         cached.
 
-    @see: L{Record_A6}
+    @see: U{http://www.faqs.org/rfcs/rfc1886.html}
     """
     implements(IEncodable, IRecord)
     TYPE = AAAA
 
+    fancybasename = 'AAAA'
+    showAttributes = (('_address', 'address', '%s'), 'ttl')
     compareAttributes = ('address', 'ttl')
 
+    _address = property(lambda self: socket.inet_ntop(AF_INET6, self.address))
+
     def __init__(self, address = '::', ttl=None):
         self.address = socket.inet_pton(AF_INET6, address)
         self.ttl = str2time(ttl)
@@ -810,15 +926,13 @@ class Record_AAAA(tputil.FancyEqMixin):               # OBSOLETE (or headed ther
         return hash(self.address)
 
 
-    def __str__(self):
-        return '<AAAA %s ttl=%s>' % (socket.inet_ntop(AF_INET6, self.address), self.ttl)
 
-
-
-class Record_A6:
+class Record_A6(tputil.FancyStrMixin, tputil.FancyEqMixin):
     """
     An IPv6 address.
 
+    This is an experimental record type.
+
     @type prefixLen: C{int}
     @ivar prefixLen: The length of the suffix.
 
@@ -837,10 +951,18 @@ class Record_A6:
         cached.
 
     @see: U{http://www.faqs.org/rfcs/rfc2874.html}
+    @see: U{http://www.faqs.org/rfcs/rfc3363.html}
+    @see: U{http://www.faqs.org/rfcs/rfc3364.html}
     """
     implements(IEncodable, IRecord)
     TYPE = A6
 
+    fancybasename = 'A6'
+    showAttributes = (('_suffix', 'suffix', '%s'), ('prefix', 'prefix', '%s'), 'ttl')
+    compareAttributes = ('prefixLen', 'prefix', 'suffix', 'ttl')
+
+    _suffix = property(lambda self: socket.inet_ntop(AF_INET6, self.suffix))
+
     def __init__(self, prefixLen=0, suffix='::', prefix='', ttl=None):
         self.prefixLen = prefixLen
         self.suffix = socket.inet_pton(AF_INET6, suffix)
@@ -873,7 +995,7 @@ class Record_A6:
                     self.suffix[-self.bytes:] == other.suffix[-self.bytes:] and
                     self.prefix == other.prefix and
                     self.ttl == other.ttl)
-        return 0
+        return NotImplemented
 
 
     def __hash__(self):
@@ -926,6 +1048,7 @@ class Record_SRV(tputil.FancyEqMixin, tputil.FancyStrMixin):
     implements(IEncodable, IRecord)
     TYPE = SRV
 
+    fancybasename = 'SRV'
     compareAttributes = ('priority', 'weight', 'target', 'port', 'ttl')
     showAttributes = ('priority', 'weight', ('target', 'target', '%s'), 'port', 'ttl')
 
@@ -955,6 +1078,100 @@ class Record_SRV(tputil.FancyEqMixin, tputil.FancyStrMixin):
 
 
 
+class Record_NAPTR(tputil.FancyEqMixin, tputil.FancyStrMixin):
+    """
+    The location of the server(s) for a specific protocol and domain.
+
+    @type order: C{int}
+    @ivar order: An integer specifying the order in which the NAPTR records
+        MUST be processed to ensure the correct ordering of rules.  Low numbers
+        are processed before high numbers.
+
+    @type preference: C{int}
+    @ivar preference: An integer that specifies the order in which NAPTR
+        records with equal "order" values SHOULD be processed, low numbers
+        being processed before high numbers.
+
+    @type flag: L{Charstr}
+    @ivar flag: A <character-string> containing flags to control aspects of the
+        rewriting and interpretation of the fields in the record.  Flags
+        aresingle characters from the set [A-Z0-9].  The case of the alphabetic
+        characters is not significant.
+
+        At this time only four flags, "S", "A", "U", and "P", are defined.
+
+    @type service: L{Charstr}
+    @ivar service: Specifies the service(s) available down this rewrite path.
+        It may also specify the particular protocol that is used to talk with a
+        service.  A protocol MUST be specified if the flags field states that
+        the NAPTR is terminal.
+
+    @type regexp: L{Charstr}
+    @ivar regexp: A STRING containing a substitution expression that is applied
+        to the original string held by the client in order to construct the
+        next domain name to lookup.
+
+    @type replacement: L{Name}
+    @ivar replacement: The next NAME to query for NAPTR, SRV, or address
+        records depending on the value of the flags field.  This MUST be a
+        fully qualified domain-name.
+
+    @type ttl: C{int}
+    @ivar ttl: The maximum number of seconds which this record should be
+        cached.
+
+    @see: U{http://www.faqs.org/rfcs/rfc2915.html}
+    """
+    implements(IEncodable, IRecord)
+    TYPE = NAPTR
+
+    compareAttributes = ('order', 'preference', 'flags', 'service', 'regexp',
+                         'replacement')
+    fancybasename = 'NAPTR'
+    showAttributes = ('order', 'preference', ('flags', 'flags', '%s'),
+                      ('service', 'service', '%s'), ('regexp', 'regexp', '%s'),
+                      ('replacement', 'replacement', '%s'), 'ttl')
+
+    def __init__(self, order=0, preference=0, flags='', service='', regexp='',
+                 replacement='', ttl=None):
+        self.order = int(order)
+        self.preference = int(preference)
+        self.flags = Charstr(flags)
+        self.service = Charstr(service)
+        self.regexp = Charstr(regexp)
+        self.replacement = Name(replacement)
+        self.ttl = str2time(ttl)
+
+
+    def encode(self, strio, compDict=None):
+        strio.write(struct.pack('!HH', self.order, self.preference))
+        # This can't be compressed
+        self.flags.encode(strio, None)
+        self.service.encode(strio, None)
+        self.regexp.encode(strio, None)
+        self.replacement.encode(strio, None)
+
+
+    def decode(self, strio, length=None):
+        r = struct.unpack('!HH', readPrecisely(strio, struct.calcsize('!HH')))
+        self.order, self.preference = r
+        self.flags = Charstr()
+        self.service = Charstr()
+        self.regexp = Charstr()
+        self.replacement = Name()
+        self.flags.decode(strio)
+        self.service.decode(strio)
+        self.regexp.decode(strio)
+        self.replacement.decode(strio)
+
+
+    def __hash__(self):
+        return hash((
+            self.order, self.preference, self.flags,
+            self.service, self.regexp, self.replacement))
+
+
+
 class Record_AFSDB(tputil.FancyStrMixin, tputil.FancyEqMixin):
     """
     Map from a domain name to the name of an AFS cell database server.
@@ -978,6 +1195,7 @@ class Record_AFSDB(tputil.FancyStrMixin, tputil.FancyEqMixin):
     implements(IEncodable, IRecord)
     TYPE = AFSDB
 
+    fancybasename = 'AFSDB'
     compareAttributes = ('subtype', 'hostname', 'ttl')
     showAttributes = ('subtype', ('hostname', 'hostname', '%s'), 'ttl')
 
@@ -1024,6 +1242,7 @@ class Record_RP(tputil.FancyEqMixin, tputil.FancyStrMixin):
     implements(IEncodable, IRecord)
     TYPE = RP
 
+    fancybasename = 'RP'
     compareAttributes = ('mbox', 'txt', 'ttl')
     showAttributes = (('mbox', 'mbox', '%s'), ('txt', 'txt', '%s'), 'ttl')
 
@@ -1050,7 +1269,7 @@ class Record_RP(tputil.FancyEqMixin, tputil.FancyStrMixin):
 
 
 
-class Record_HINFO(tputil.FancyStrMixin):
+class Record_HINFO(tputil.FancyStrMixin, tputil.FancyEqMixin):
     """
     Host information.
 
@@ -1067,7 +1286,8 @@ class Record_HINFO(tputil.FancyStrMixin):
     implements(IEncodable, IRecord)
     TYPE = HINFO
 
-    showAttributes = ('cpu', 'os', 'ttl')
+    fancybasename = 'HINFO'
+    showAttributes = compareAttributes = ('cpu', 'os', 'ttl')
 
     def __init__(self, cpu='', os='', ttl=None):
         self.cpu, self.os = cpu, os
@@ -1091,7 +1311,7 @@ class Record_HINFO(tputil.FancyStrMixin):
             return (self.os.lower() == other.os.lower() and
                     self.cpu.lower() == other.cpu.lower() and
                     self.ttl == other.ttl)
-        return 0
+        return NotImplemented
 
 
     def __hash__(self):
@@ -1126,6 +1346,7 @@ class Record_MINFO(tputil.FancyEqMixin, tputil.FancyStrMixin):
     rmailbx = None
     emailbx = None
 
+    fancybasename = 'MINFO'
     compareAttributes = ('rmailbx', 'emailbx', 'ttl')
     showAttributes = (('rmailbx', 'responsibility', '%s'),
                       ('emailbx', 'errors', '%s'),
@@ -1171,6 +1392,7 @@ class Record_MX(tputil.FancyStrMixin, tputil.FancyEqMixin):
     implements(IEncodable, IRecord)
     TYPE = MX
 
+    fancybasename = 'MX'
     compareAttributes = ('preference', 'name', 'ttl')
     showAttributes = ('preference', ('name', 'name', '%s'), 'ttl')
 
@@ -1206,11 +1428,15 @@ class Record_TXT(tputil.FancyEqMixin, tputil.FancyStrMixin):
 
     @type data: C{list} of C{str}
     @ivar data: Freeform text which makes up this record.
+
+    @type ttl: C{int}
+    @ivar ttl: The maximum number of seconds which this record should be cached.
     """
     implements(IEncodable, IRecord)
 
     TYPE = TXT
 
+    fancybasename = 'TXT'
     showAttributes = compareAttributes = ('data', 'ttl')
 
     def __init__(self, *data, **kw):
@@ -1233,8 +1459,8 @@ class Record_TXT(tputil.FancyEqMixin, tputil.FancyStrMixin):
             soFar += L + 1
         if soFar != length:
             log.msg(
-                "Decoded %d bytes in TXT record, but rdlength is %d" % (
-                    soFar, length
+                "Decoded %d bytes in %s record, but rdlength is %d" % (
+                    soFar, self.fancybasename, length
                 )
             )
 
@@ -1244,7 +1470,75 @@ class Record_TXT(tputil.FancyEqMixin, tputil.FancyStrMixin):
 
 
 
+# This is a fallback record
+class UnknownRecord(tputil.FancyEqMixin, tputil.FancyStrMixin, object):
+    """
+    Encapsulate the wire data for unkown record types so that they can
+    pass through the system unchanged.
+
+    @type data: C{str}
+    @ivar data: Wire data which makes up this record.
+
+    @type ttl: C{int}
+    @ivar ttl: The maximum number of seconds which this record should be cached.
+
+    @since: 11.1
+    """
+    implements(IEncodable, IRecord)
+
+    fancybasename = 'UNKNOWN'
+    compareAttributes = ('data', 'ttl')
+    showAttributes = ('data', 'ttl')
+
+    def __init__(self, data='', ttl=None):
+        self.data = data
+        self.ttl = str2time(ttl)
+
+
+    def encode(self, strio, compDict=None):
+        """
+        Write the raw bytes corresponding to this record's payload to the
+        stream.
+        """
+        strio.write(self.data)
+
+
+    def decode(self, strio, length=None):
+        """
+        Load the bytes which are part of this record from the stream and store
+        them unparsed and unmodified.
+        """
+        if length is None:
+            raise Exception('must know length for unknown record types')
+        self.data = readPrecisely(strio, length)
+
+
+    def __hash__(self):
+        return hash((self.data, self.ttl))
+
+
+
+class Record_SPF(Record_TXT):
+    """
+    Structurally, freeform text. Semantically, a policy definition, formatted
+    as defined in U{rfc 4408<http://www.faqs.org/rfcs/rfc4408.html>}.
+
+    @type data: C{list} of C{str}
+    @ivar data: Freeform text which makes up this record.
+
+    @type ttl: C{int}
+    @ivar ttl: The maximum number of seconds which this record should be cached.
+    """
+    TYPE = SPF
+    fancybasename = 'SPF'
+
+
+
 class Message:
+    """
+    L{Message} contains all the information represented by a single
+    DNS request or response.
+    """
     headerFmt = "!H2B4H"
     headerSize = struct.calcsize(headerFmt)
 
@@ -1343,7 +1637,7 @@ class Message:
 
     def parseRecords(self, list, num, strio):
         for i in range(num):
-            header = RRHeader()
+            header = RRHeader(auth=self.auth)
             try:
                 header.decode(strio)
             except EOFError:
@@ -1359,8 +1653,32 @@ class Message:
             list.append(header)
 
 
+    # Create a mapping from record types to their corresponding Record_*
+    # classes.  This relies on the global state which has been created so
+    # far in initializing this module (so don't define Record classes after
+    # this).
+    _recordTypes = {}
+    for name in globals():
+        if name.startswith('Record_'):
+            _recordTypes[globals()[name].TYPE] = globals()[name]
+
+    # Clear the iteration variable out of the class namespace so it
+    # doesn't become an attribute.
+    del name
+
+
     def lookupRecordType(self, type):
-        return globals().get('Record_' + QUERY_TYPES.get(type, ''), None)
+        """
+        Retrieve the L{IRecord} implementation for the given record type.
+
+        @param type: A record type, such as L{A} or L{NS}.
+        @type type: C{int}
+
+        @return: An object which implements L{IRecord} or C{None} if none
+            can be found for the given type.
+        @rtype: L{types.ClassType}
+        """
+        return self._recordTypes.get(type, UnknownRecord)
 
 
     def toStr(self):
@@ -1373,34 +1691,42 @@ class Message:
         strio = StringIO.StringIO(str)
         self.decode(strio)
 
+
+
 class DNSMixin(object):
     """
     DNS protocol mixin shared by UDP and TCP implementations.
+
+    @ivar _reactor: A L{IReactorTime} and L{IReactorUDP} provider which will
+        be used to issue DNS queries and manage request timeouts.
     """
     id = None
     liveMessages = None
 
-    def __init__(self, controller):
+    def __init__(self, controller, reactor=None):
         self.controller = controller
         self.id = random.randrange(2 ** 10, 2 ** 15)
+        if reactor is None:
+            from twisted.internet import reactor
+        self._reactor = reactor
+
 
     def pickID(self):
         """
         Return a unique ID for queries.
         """
         while True:
-            self.id += randomSource() % (2 ** 10)
-            self.id %= 2 ** 16
-            if self.id not in self.liveMessages:
-                break
-        return self.id
+            id = randomSource()
+            if id not in self.liveMessages:
+                return id
+
 
     def callLater(self, period, func, *args):
         """
         Wrapper around reactor.callLater, mainly for test purpose.
         """
-        from twisted.internet import reactor
-        return reactor.callLater(period, func, *args)
+        return self._reactor.callLater(period, func, *args)
+
 
     def _query(self, queries, timeout, id, writeMessage):
         """
@@ -1478,8 +1804,7 @@ class DNSDatagramProtocol(DNSMixin, protocol.DatagramProtocol):
         self.transport.write(message.toStr(), address)
 
     def startListening(self):
-        from twisted.internet import reactor
-        reactor.listenUDP(0, self, maxPacketSize=512)
+        self._reactor.listenUDP(0, self, maxPacketSize=512)
 
     def datagramReceived(self, data, addr):
         """
@@ -1576,6 +1901,15 @@ class DNSProtocol(DNSMixin, protocol.Protocol):
         self.liveMessages = {}
         self.controller.connectionMade(self)
 
+
+    def connectionLost(self, reason):
+        """
+        Notify the controller that this protocol is no longer
+        connected.
+        """
+        self.controller.connectionLost(self)
+
+
     def dataReceived(self, data):
         self.buffer += data
 
@@ -1607,6 +1941,7 @@ class DNSProtocol(DNSMixin, protocol.Protocol):
             else:
                 break
 
+
     def query(self, queries, timeout=60):
         """
         Send out a message with the given queries.
@@ -1618,4 +1953,3 @@ class DNSProtocol(DNSMixin, protocol.Protocol):
         """
         id = self.pickID()
         return self._query(queries, timeout, id, self.writeMessage)
-
